﻿using System;
using System.ComponentModel;
using System.IO;
using System.Reflection;
using System.Collections.Generic;
using System.Xml;
using System.Xml.Serialization;

namespace ScpControl 
{
    public partial class BusDevice : ScpDevice
    {
        #region default xml file
        string defaultfile = @"<Profile>
  <button>
    <name>up</name>
    <mode>Xbox Input Mode</mode>
    <value>Up</value>
  </button>
  <button>
    <name>down</name>
    <mode>Xbox Input Mode</mode>
    <value>Down</value>
  </button>
  <button>
    <name>left</name>
    <mode>Xbox Input Mode</mode>
    <value>Left</value>
  </button>
  <button>
    <name>right</name>
    <mode>Xbox Input Mode</mode>
    <value>Right</value>
  </button>
  <button>
    <name>start</name>
    <mode>Xbox Input Mode</mode>
    <value>Start</value>
  </button>
  <button>
    <name>select</name>
    <mode>Xbox Input Mode</mode>
    <value>Back</value>
  </button>
  <button>
    <name>l1</name>
    <mode>Xbox Input Mode</mode>
    <value>LB</value>
  </button>
  <button>
    <name>l2</name>
    <mode>Xbox Input Mode</mode>
    <value>LT</value>
  </button>
  <button>
    <name>l3</name>
    <mode>Xbox Input Mode</mode>
    <value>LS</value>
  </button>
  <button>
    <name>r1</name>
    <mode>Xbox Input Mode</mode>
    <value>RB</value>
  </button>
  <button>
    <name>r2</name>
    <mode>Xbox Input Mode</mode>
    <value>RT</value>
  </button>
  <button>
    <name>r3</name>
    <mode>Xbox Input Mode</mode>
    <value>RS</value>
  </button>
  <button>
    <name>cross</name>
    <mode>Xbox Input Mode</mode>
    <value>A</value>
  </button>
  <button>
    <name>circle</name>
    <mode>Xbox Input Mode</mode>
    <value>B</value>
  </button>
  <button>
    <name>square</name>
    <mode>Xbox Input Mode</mode>
    <value>X</value>
  </button>
  <button>
    <name>triangle</name>
    <mode>Xbox Input Mode</mode>
    <value>Y</value>
  </button>
  <digital_triggers>
    <value>False</value>
  </digital_triggers>
  <axis>
    <name>X-Axis</name>
    <value>LX</value>
  </axis>
  <axis>
    <name>Y-Axis</name>
    <value>LY</value>
  </axis>
  <axis>
    <name>RX-Axis</name>
    <value>RX</value>
  </axis>
  <axis>
    <name>RY-Axis</name>
    <value>RY</value>
  </axis>
</Profile>";
        #endregion

        #region default settings file
        string defaultsettingsfile = @"<Settings>
  <ActiveProfile>default</ActiveProfile>
</Settings>"; 
        #endregion

        public const String SCP_BUS_CLASS_GUID = "{F679F562-3164-42CE-A4DB-E7DDBE723909}";
        public const Int32 ReportSize = 28;
        public const Int32 RumbleSize =  8;
        public const Int32 BusWidth   =  4;

        protected DsState m_State = DsState.Disconnected;
        public DsState State 
        {
            get { return m_State; }
        }

        protected Int32 m_Offset = 0;
        protected List<Int32> m_Plugged = new List<Int32>();

        protected virtual Int32 IndexToSerial(Byte Index) 
        {
            return Index + m_Offset + 1;
        }

        public event EventHandler<DebugEventArgs> Debug = null;

        protected virtual void LogDebug(String Data) 
        {
            DebugEventArgs args = new DebugEventArgs(Data);

            if (Debug != null)
            {
                Debug(this, args);
            }
        }


        protected virtual Int32 Scale(Int32 Value, Boolean Flip) 
        {
            Value -= 0x80; if (Value == -128) Value = -127;

            if (Flip) Value *= -1;

            return (Int32)((float) Value * 258.00787401574803149606299212599f);
        }

        protected virtual Boolean DeadZone(Int32 R, Int32 X, Int32 Y) 
        {
            X -= 0x80; if (X == -128) X = -127;
            Y -= 0x80; if (Y == -128) Y = -127;

            return R * R >= X * X + Y * Y;
        }


        public BusDevice() : base(SCP_BUS_CLASS_GUID) 
        {
            InitializeComponent();
        }

        public BusDevice(IContainer container) : base(SCP_BUS_CLASS_GUID) 
        {
            container.Add(this);

            InitializeComponent();
        }


        public override Boolean Open(Int32 Instance = 0) 
        {
            if (State == DsState.Disconnected)
            {
                m_Offset = Instance * BusWidth;

                LogDebug(String.Format("-- Bus Open   : Offset {0}", m_Offset));

                if (!base.Open(0))
                {
                    LogDebug(String.Format("-- Bus Open   : Failed!!", m_Offset));
                }
            }

            return State == DsState.Reserved;
        }

        public override Boolean Open(String DevicePath)  
        {
            if (State == DsState.Disconnected)
            {
                m_Path = DevicePath;

                LogDebug(String.Format("-- Bus Open   : Path {0}", m_Path));

                if (GetDeviceHandle(m_Path))
                {
                    m_IsActive = true;
                    m_State = DsState.Reserved;
                }
            }

            return State == DsState.Reserved;
        }

        public override Boolean Start() 
        {
            if (State == DsState.Reserved)
            {
                m_State = DsState.Connected;
            }

            return State == DsState.Connected;
        }

        public override Boolean Stop()  
        {
            if (State == DsState.Connected)
            {
                Queue<Int32> Items = new Queue<Int32>();

                lock (m_Plugged)
                {
                    foreach (Int32 Serial in m_Plugged) Items.Enqueue(Serial - m_Offset);
                }

                while (Items.Count > 0) Unplug(Items.Dequeue());

                m_State = DsState.Reserved;
            }

            return State == DsState.Reserved;
        }

        public override Boolean Close() 
        {
            if (base.Stop())
            {
                m_State = DsState.Reserved;
            }

            if (State != DsState.Reserved)
            {
                if (base.Close())
                {
                    m_State = DsState.Disconnected;
                }
            }

            return State == DsState.Disconnected;
        }


        public virtual Boolean Suspend() 
        {
            return Stop();
        }

        public virtual Boolean Resume()  
        {
            return Start();
            
        }

        

        public virtual Int32 Parse(Byte[] Input, Byte[] Output, DsModel Type = DsModel.DS3) 
        {
            Int32 Serial = IndexToSerial(Input[0]);

            bool FlipLX, FlipLY, FlipRX, FlipRY;
            Int32 RX_axis=0, RY_axis=0, LX_axis=0, LY_axis=0;
            Int32 tilt_offsetLX = 0;
            Int32 tilt_offsetLY = 0;
            Int32 tilt_offsetRX = 0;
            Int32 tilt_offsetRY = 0;

            string activeprofile_filename = AppDomain.CurrentDomain.BaseDirectory + "Settings.xml";
            XmlDocument activeprofile_xml = new XmlDocument();
            if (!File.Exists(activeprofile_filename)) File.WriteAllText(activeprofile_filename,defaultsettingsfile);
            activeprofile_xml.Load(activeprofile_filename);
            XmlNode profile_name = activeprofile_xml.SelectSingleNode("//Settings/ActiveProfile");


            string filename = AppDomain.CurrentDomain.BaseDirectory + profile_name.InnerText + ".xml";
            XmlDocument xmlDoc = new XmlDocument();
            if (!File.Exists(filename)) File.WriteAllText(filename, defaultfile);
            xmlDoc.Load(filename);
            XmlNodeList mainlist = xmlDoc.SelectNodes("//Profile/button");
            
            XmlNodeList axislist = xmlDoc.SelectNodes("//Profile/axis");
            
            XmlNode triggerstatus = xmlDoc.SelectSingleNode("//Profile/digital_triggers/value");
            XmlNode mode, value;
            bool trigger = Convert.ToBoolean(triggerstatus.InnerText);
            
            LX_axis = axisinfo(axislist[0].SelectSingleNode("value").InnerText);
            LY_axis = axisinfo(axislist[1].SelectSingleNode("value").InnerText);
            RX_axis = axisinfo(axislist[2].SelectSingleNode("value").InnerText);
            RY_axis = axisinfo(axislist[3].SelectSingleNode("value").InnerText);

            if (LX_axis == 50 || LX_axis == 52)
            {
                tilt_offsetLX = 0x80;
                FlipLX = true;
            }
            else
            {
                tilt_offsetLX = 0;
                FlipLX = false;
            }

            if (LY_axis == 50 || LY_axis == 52)
            {
                tilt_offsetLY = 0x80;
                FlipLY = true;
            }
            else
            {
                tilt_offsetLY = 0;
                FlipLY = false;
            }

            if (RX_axis == 50 || RX_axis == 52)
            {
                tilt_offsetRX = 0x80;
                FlipRX = true;
            }
            else
            {
                tilt_offsetRX = 0;
                FlipRX = false;
            }

            if (RY_axis == 50 || RY_axis == 52)
            {
                tilt_offsetRY = 0x80;
                FlipRY = true;
            }
            else
            {
                tilt_offsetRY = 0;
                FlipRY = false;
            }

            for (Int32 Index = 0; Index < ReportSize; Index++) Output[Index] = 0x00;

            Output[0] = 0x1C;
            Output[4] = (Byte)((Serial >>  0) & 0xFF);
            Output[5] = (Byte)((Serial >>  8) & 0xFF);
            Output[6] = (Byte)((Serial >> 16) & 0xFF);
            Output[7] = (Byte)((Serial >> 24) & 0xFF);
            Output[9] = 0x14;

            X360Button XButton = X360Button.None;
            
            if (Input[1] == 0x02) // Pad is active
            {
                switch (Type)
                {
                    case DsModel.DS3:
                        {
                            Ds3Button Buttons = (Ds3Button)((Input[10] << 0) | (Input[11] << 8) | (Input[12] << 16) | (Input[13] << 24));

                            mode = mainlist[0].SelectSingleNode("mode"); value = mainlist[0].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Up, Ds3Axis.Up, ref XButton, ref Input, ref Output,mode.InnerText,value.InnerText,trigger);
                            mode = mainlist[1].SelectSingleNode("mode"); value = mainlist[1].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Down, Ds3Axis.Down, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[2].SelectSingleNode("mode"); value = mainlist[2].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Left, Ds3Axis.Left, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[3].SelectSingleNode("mode"); value = mainlist[3].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Right, Ds3Axis.Right, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);

                            mode = mainlist[4].SelectSingleNode("mode"); value = mainlist[4].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Start, Ds3Axis.None, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[5].SelectSingleNode("mode"); value = mainlist[5].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Select, Ds3Axis.None, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);

                            mode = mainlist[6].SelectSingleNode("mode"); value = mainlist[6].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.L1, Ds3Axis.L1, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[7].SelectSingleNode("mode"); value = mainlist[7].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.L2, Ds3Axis.L2, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[8].SelectSingleNode("mode"); value = mainlist[8].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.L3, Ds3Axis.None, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);

                            mode = mainlist[9].SelectSingleNode("mode"); value = mainlist[9].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.R1, Ds3Axis.R1, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[10].SelectSingleNode("mode"); value = mainlist[10].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.R2, Ds3Axis.R2, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[11].SelectSingleNode("mode"); value = mainlist[11].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.R3, Ds3Axis.None, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);

                            mode = mainlist[12].SelectSingleNode("mode"); value = mainlist[12].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Cross, Ds3Axis.Cross, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[13].SelectSingleNode("mode"); value = mainlist[13].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Circle, Ds3Axis.Circle, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[14].SelectSingleNode("mode"); value = mainlist[14].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Square, Ds3Axis.Square, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            mode = mainlist[15].SelectSingleNode("mode"); value = mainlist[15].SelectSingleNode("value");
                            input(ref Buttons, Ds3Button.Triangle, Ds3Axis.Triangle, ref XButton, ref Input, ref Output, mode.InnerText, value.InnerText, trigger);
                            
                            //ps button is preffered no to be re-assigned as it may break compatibility with the ds3 control center
                            if (Buttons.HasFlag(Ds3Button.PS)) XButton |= X360Button.Guide;
                            ///////////////////////////////////////////////////////////////////////////////////////////////////////

                            //send the buttons data to xbox360 device
                            Output[(UInt32) X360Axis.BT_Lo] = (Byte)((UInt32) XButton >> 0 & 0xFF);
                            Output[(UInt32) X360Axis.BT_Hi] = (Byte)((UInt32) XButton >> 8 & 0xFF);

                            if (!DeadZone((byte)10, Input[(UInt32) LX_axis], Input[(UInt32) LY_axis]))   // Left Stick DeadZone
                            {
                                Int32 ThumbLX = +Scale((Input[(UInt32)LX_axis] + tilt_offsetLX), FlipLX);
                                Int32 ThumbLY = -Scale((Input[(UInt32)LY_axis] + tilt_offsetLY), FlipLY);

                                Output[(UInt32) X360Axis.LX_Lo] = (Byte)((ThumbLX >> 0) & 0xFF); // LX
                                Output[(UInt32) X360Axis.LX_Hi] = (Byte)((ThumbLX >> 8) & 0xFF);

                                Output[(UInt32) X360Axis.LY_Lo] = (Byte)((ThumbLY >> 0) & 0xFF); // LY
                                Output[(UInt32) X360Axis.LY_Hi] = (Byte)((ThumbLY >> 8) & 0xFF);
                            }

                            if (!DeadZone((byte)10, Input[(UInt32) RX_axis], Input[(UInt32)RY_axis]))   // Right Stick DeadZone
                            {
                                Int32 ThumbRX = +Scale((Input[(UInt32)RX_axis] + tilt_offsetRX), FlipRX);
                                Int32 ThumbRY = -Scale((Input[(UInt32)RY_axis] + tilt_offsetRY), FlipRY);

                                Output[(UInt32) X360Axis.RX_Lo] = (Byte)((ThumbRX >> 0) & 0xFF); // RX
                                Output[(UInt32) X360Axis.RX_Hi] = (Byte)((ThumbRX >> 8) & 0xFF);

                                Output[(UInt32) X360Axis.RY_Lo] = (Byte)((ThumbRY >> 0) & 0xFF); // RY
                                Output[(UInt32) X360Axis.RY_Hi] = (Byte)((ThumbRY >> 8) & 0xFF);
                            }
                        }
                        break;
                }
            }

            return Input[0];
        }


        public virtual Boolean Plugin(Int32 Serial) 
        {
            Boolean retVal = false;

            if (Serial < 1 || Serial > BusWidth) return retVal;

            Serial += m_Offset;

            if (State == DsState.Connected)
            {
                lock (m_Plugged)
                {
                    if (!m_Plugged.Contains(Serial))
                    {
                        Int32 Transfered = 0;
                        Byte[] Buffer = new Byte[16];

                        Buffer[0] = 0x10;
                        Buffer[1] = 0x00;
                        Buffer[2] = 0x00;
                        Buffer[3] = 0x00;

                        Buffer[4] = (Byte)((Serial >>  0) & 0xFF);
                        Buffer[5] = (Byte)((Serial >>  8) & 0xFF);
                        Buffer[6] = (Byte)((Serial >> 16) & 0xFF);
                        Buffer[7] = (Byte)((Serial >> 24) & 0xFF);

                        if (DeviceIoControl(m_FileHandle, 0x2A4000, Buffer, Buffer.Length, null, 0, ref Transfered, IntPtr.Zero))
                        {
                            m_Plugged.Add(Serial); retVal = true;

                            LogDebug(String.Format("-- Bus Plugin : Serial {0}", Serial));
                        }
                    }
                    else retVal = true;
                }
            }

            return retVal;
        }

        public virtual Boolean Unplug(Int32 Serial) 
        {
            Boolean retVal = false;
            Serial += m_Offset;

            if (State == DsState.Connected)
            {
                lock (m_Plugged)
                {
                    if (m_Plugged.Contains(Serial))
                    {
                        Int32 Transfered = 0;
                        Byte[] Buffer = new Byte[16];

                        Buffer[0] = 0x10;
                        Buffer[1] = 0x00;
                        Buffer[2] = 0x00;
                        Buffer[3] = 0x00;

                        Buffer[4] = (Byte)((Serial >>  0) & 0xFF);
                        Buffer[5] = (Byte)((Serial >>  8) & 0xFF);
                        Buffer[6] = (Byte)((Serial >> 16) & 0xFF);
                        Buffer[7] = (Byte)((Serial >> 24) & 0xFF);

                        if (DeviceIoControl(m_FileHandle, 0x2A4004, Buffer, Buffer.Length, null, 0, ref Transfered, IntPtr.Zero))
                        {
                            m_Plugged.Remove(Serial); retVal = true;

                            LogDebug(String.Format("-- Bus Unplug : Serial {0}", Serial));
                        }
                    }
                    else retVal = true;
                }
            }

            return retVal;
        }

        public virtual Boolean Report(Byte[] Input, Byte[] Output) 
        {
            if (State == DsState.Connected)
            {
                Int32 Transfered = 0;

                return DeviceIoControl(m_FileHandle, 0x2A400C, Input, Input.Length, Output, Output.Length, ref Transfered, IntPtr.Zero) && Transfered > 0;
            }

            return false;
        }

        public void input(ref Ds3Button buttons,Ds3Button btn,Ds3Axis axs,ref X360Button button_x360,ref byte[] Input,ref byte[] Output,string mode,string value,bool triggervalue)
        {
            if (mode == "Xbox Input Mode")
            {
                switch (value)
                {
                    case "Up": if (buttons.HasFlag(btn)) button_x360 |= X360Button.Up;
                        break;
                    case "Down": if (buttons.HasFlag(btn)) button_x360 |= X360Button.Down;
                        break;
                    case "Left": if (buttons.HasFlag(btn)) button_x360 |= X360Button.Left;
                        break;
                    case "Right": if (buttons.HasFlag(btn)) button_x360 |= X360Button.Right;
                        break;
                    case "Start": if (buttons.HasFlag(btn)) button_x360 |= X360Button.Start;
                        break;
                    case "Back": if (buttons.HasFlag(btn)) button_x360 |= X360Button.Back;
                        break;
                    case "LB": if (buttons.HasFlag(btn)) button_x360 |= X360Button.LB;
                        break;
                    case "LT": switch (btn)
                        {
                            case Ds3Button.Select:
                            case Ds3Button.Start:
                            case Ds3Button.L3:
                            case Ds3Button.R3:
                                Output[(UInt32)X360Axis.LT] = (byte)(buttons.HasFlag(btn)? 255:0);
                                break;
                            default: if (triggervalue)
                                    Output[(UInt32)X360Axis.LT] = (byte)(buttons.HasFlag(btn) ? 255 : 0);
                                else
                                    Output[(UInt32)X360Axis.LT] = Input[(UInt32)axs];
                                break;
                        }
                        break;
                    case "LS": if (buttons.HasFlag(btn)) button_x360 |= X360Button.LS;
                        break;
                    case "RB": if (buttons.HasFlag(btn)) button_x360 |= X360Button.RB;
                        break;
                    case "RT": switch (btn)
                        {
                            case Ds3Button.Select:
                            case Ds3Button.Start:
                            case Ds3Button.L3:
                            case Ds3Button.R3:
                                Output[(UInt32)X360Axis.RT] = (byte)(buttons.HasFlag(btn) ? 255 : 0);
                                break;
                            default: if (triggervalue)
                                    Output[(UInt32)X360Axis.RT] = (byte)(buttons.HasFlag(btn) ? 255 : 0);
                                else
                                    Output[(UInt32)X360Axis.RT] = Input[(UInt32)axs];
                                break;
                        }
                        break;
                    case "RS": if (buttons.HasFlag(btn)) button_x360 |= X360Button.RS;
                        break;
                    case "A": if (buttons.HasFlag(btn)) button_x360 |= X360Button.A;
                        break;
                    case "B": if (buttons.HasFlag(btn)) button_x360 |= X360Button.B;
                        break;
                    case "X": if (buttons.HasFlag(btn)) button_x360 |= X360Button.X;
                        break;
                    case "Y": if (buttons.HasFlag(btn)) button_x360 |= X360Button.Y;
                        break;
                }
            }
        }

        public int axisinfo(string axis)
        {
            int value = 0;
            switch (axis)
            {
                case "LX": value = 14;
                    break;
                case "LY": value = 15;
                    break;
                case "RX": value = 16;
                    break;
                case "RY": value = 17;
                    break;
                case "Tilt-X": value = 50;
                    break;
                case "Tilt-Y": value = 52;
                    break;
                case "Tilt-Z": value = 54;
                    break;
            }
            return value;
        }
    }
}
